@@ -67,45 +67,60 @@ The provided schema can be cut/pasted or fed directly into the client program
 to create a database. The '--' introduces comments according to SQL specs.
 
 -- MySQL notes:
---   - the attribute SERIAL was introduced with MySQL 4.1.0;
---     with earlier versions one can use INT UNSIGNED NOT NULL AUTO_INCREMENT
+--   - SERIAL can be used instead of INT UNSIGNED NOT NULL AUTO_INCREMENT
+--     with databases which do not recognize AUTO_INCREMENT;
+--     The attribute SERIAL was introduced with MySQL 4.1.0, but it
+--     implicitly creates an additional UNIQUE index, which is redundant.
+--   - instead of declaring a time_iso field in table msgs as a string:
+--         time_iso char(16) NOT NULL,
+--       it is may be more useful to declare is as:
+--         time_iso TIMESTAMP,
+--       and set $timestamp_fmt_mysql=1 in amavisd.conf to avoid MySQL
+--       inability to accept ISO 8601 timestamps with zone Z and delimiter T
 
 -- PostgreSQL notes (initially provided by Phil Regnauld):
+--   - use SERIAL instead of INT UNSIGNED NOT NULL AUTO_INCREMENT
 --   - remove the 'unsigned' throughout,
+--   - remove the 'ENGINE=InnoDB' throughout,
+--   - instead of declaring a time_iso field in table msgs as a string:
+--         time_iso char(16) NOT NULL,
+--       it is more useful to declare is as:
+--         time_iso TIMESTAMP WITHOUT TIME ZONE NOT NULL,
+--       if changing existing table from char to timestamp is desired,
+--       the following clause can be used:
+--         ALTER TABLE msgs ALTER time_iso
+--           TYPE TIMESTAMP WITH TIME ZONE
+--           USING to_timestamp(time_iso,'YYYYMMDDTHH24MMSSTZ');
 --   - create an amavis username and the database (choose name, e.g. mail)
---       $ creatuser -U pgsql --no-adduser --createdb amavis
+--       $ createuser -U pgsql --no-adduser --createdb amavis
 --       $ createdb -U amavis mail
 --   - populate the database using the schema below:
 --       $ psql -U amavis mail < amavisd-pg.sql
---   - leave out 'KEY ...'
 
 -- SQLite notes:
+--   - use INTEGER PRIMARY KEY AUTOINCREMENT
+--     instead of INT UNSIGNED NOT NULL AUTO_INCREMENT
 --   - SQLite is well suited for lookups database, but is not appropriate
 --     for @storage_sql_dsn due to coarse lock granularity;
 --   - replace SERIAL by INTEGER,
---   - leave out 'KEY ...'
 
 -- local users
 CREATE TABLE users (
-  id         SERIAL PRIMARY KEY,  -- unique id, not necessarily auto-assigned
+  id         INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,  -- unique id
   priority   integer      NOT NULL DEFAULT '7',  -- sort field, 0 is low prior.
   policy_id  integer unsigned NOT NULL DEFAULT '1',  -- JOINs with policy.id
-  email      varchar(255) NOT NULL,
+  email      varchar(255) NOT NULL UNIQUE,
   fullname   varchar(255) DEFAULT NULL,    -- not used by amavisd-new
-  local      char(1),     -- Y/N  (optional field, see note further down)
-  KEY email (email)
+  local      char(1)      -- Y/N  (optional field, see note further down)
 );
-CREATE UNIQUE INDEX users_idx_email ON users (email);
 
 -- any e-mail address (non- rfc2822-quoted), external or local,
 -- used as senders in wblist
 CREATE TABLE mailaddr (
-  id         SERIAL PRIMARY KEY,
+  id         INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
   priority   integer      NOT NULL DEFAULT '7',  -- 0 is low priority
-  email      varchar(255) NOT NULL,
-  KEY email (email)
+  email      varchar(255) NOT NULL UNIQUE
 );
-CREATE UNIQUE INDEX mailaddr_idx_email ON mailaddr (email);
 
 -- per-recipient whitelist and/or blacklist,
 -- puts sender and recipient in relation wb  (white or blacklisted sender)
@@ -117,8 +132,9 @@ CREATE TABLE wblist (
 );
 
 CREATE TABLE policy (
-  id         SERIAL PRIMARY KEY,    -- this is the _only_ required field
-  policy_name      varchar(32),     -- not used by amavisd-new
+  id  INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
+                                    -- 'id' this is the _only_ required field
+  policy_name      varchar(32),     -- not used by amavisd-new, a comment
 
   virus_lover          char(1) default NULL,     -- Y/N
   spam_lover           char(1) default NULL,     -- Y/N
@@ -137,10 +153,10 @@ CREATE TABLE policy (
   banned_quarantine_to     varchar(64) default NULL,
   bad_header_quarantine_to varchar(64) default NULL,
 
-  spam_tag_level  float default NULL,  -- higher score inserts spam info headers
-  spam_tag2_level float default NULL,  -- inserts 'declared spam' header fields
-  spam_kill_level float default NULL,  -- higher score activates evasive actions, e.g.
-                                       -- reject/drop, quarantine, ...
+  spam_tag_level  float default NULL, -- higher score inserts spam info headers
+  spam_tag2_level float default NULL, -- inserts 'declared spam' header fields
+  spam_kill_level float default NULL, -- higher score triggers evasive actions
+                                      -- e.g. reject/drop, quarantine, ...
                                      -- (subject to final_spam_destiny setting)
   spam_dsn_cutoff_level float default NULL,
 
@@ -168,7 +184,7 @@ CREATE TABLE policy (
 
 -- R/W part of the dataset (optional)
 --   May reside in the same or in a separate database as lookups database;
---   requires support for transactions; specified in @storage_sql_dsn
+--   REQUIRES SUPPORT FOR TRANSACTIONS; specified in @storage_sql_dsn
 --
 --   MySQL note ( http://dev.mysql.com/doc/mysql/en/storage-engines.html ):
 --     ENGINE is the preferred term, but cannot be used before MySQL 4.0.18.
@@ -188,22 +204,25 @@ CREATE TABLE policy (
 
 -- provide unique id for each e-mail address, avoids storing copies
 CREATE TABLE maddr (
-  id         SERIAL PRIMARY KEY,
-  email      varchar(255) NOT NULL,    -- full mail address
+  id         INT UNSIGNED NOT NULL AUTO_INCREMENT PRIMARY KEY,
+  email      varchar(255) NOT NULL UNIQUE, -- full mail address
   domain     varchar(255) NOT NULL     -- only domain part of the email address
                                        -- with subdomain fields in reverse
 ) ENGINE=InnoDB;
-CREATE UNIQUE INDEX maddr_idx_email     ON maddr (email);
-CREATE        INDEX maddr_idx_domain    ON maddr (domain);
 
 -- information pertaining to each processed message as a whole;
 -- NOTE: records with NULL msgs.content should be ignored by utilities,
 --   as such records correspond to messages just being processes, or were lost
+-- NOTE: with PostgreSQL, instead of a character field time_iso, please use:
+--   time_iso timestamp without time zone NOT NULL,
+-- NOTE: with MySQL, instead of a character field time_iso, one might prefer:
+--   time_iso timestamp NOT NULL,
+--   but a change in the program is required: edit: $timestamp_fmt_mysql=1
 CREATE TABLE msgs (
-  mail_id    varchar(12)   NOT NULL,    -- long-term unique mail id
+  mail_id    varchar(12)   NOT NULL PRIMARY KEY,  -- long-term unique mail id
   secret_id  varchar(12)   DEFAULT '',  -- authorizes release of mail_id
   am_id      varchar(20)   NOT NULL,    -- id used in the log
-  time_num   integer unsigned NOT NULL, -- rx_time: second since Unix epoch
+  time_num   integer unsigned NOT NULL, -- rx_time: seconds since Unix epoch
   time_iso   char(16)      NOT NULL,    -- rx_time: ISO8601 UTC ascii time
   sid        integer unsigned NOT NULL, -- sender: maddr.id
   policy     varchar(255)  DEFAULT '',  -- policy bank path (like macro %p)
@@ -213,20 +232,24 @@ CREATE TABLE msgs (
                                         -- ...on partially processed mail
   quar_type  char(1),                   -- quarantined as: ' '/F/Z/B/Q/M
                                         --  none/file/zipfile/bsmtp/sql/mailbox
+  quar_loc   varchar(255)  DEFAULT '',  -- quarantine location (e.g. file)
   dsn_sent   char(1),                   -- was DSN sent? Y/N/q (q=quenched)
   spam_level float,                     -- base message spam level (no boosts)
   message_id varchar(255)  DEFAULT '',  -- mail Message-ID header field
   from_addr  varchar(255)  DEFAULT '',  -- mail From header field,    UTF8
   subject    varchar(255)  DEFAULT '',  -- mail Subject header field, UTF8
   host       varchar(255)  NOT NULL,    -- hostname where amavisd is running
-  PRIMARY KEY (mail_id),
-  KEY sid (sid)
+  FOREIGN KEY (sid) REFERENCES maddr(id) ON DELETE RESTRICT
 ) ENGINE=InnoDB;
-CREATE INDEX msgs_idx_sid ON msgs (sid);
+CREATE INDEX msgs_idx_sid      ON msgs (sid);
+CREATE INDEX msgs_idx_time_num ON msgs (time_num);
+-- alternatively when purging based on time_iso (instead of msgs_idx_time_num):
+-- CREATE INDEX msgs_idx_time_iso ON msgs (time_iso);
 
 -- per-recipient information related to each processed message;
 -- NOTE: records in msgrcpt without corresponding msgs.mail_id record are
 --  orphaned and should be ignored and eventually deleted by external utilities
+-- NOTE: time_num field was introduced to this table with version 2.4.0
 CREATE TABLE msgrcpt (
   mail_id    varchar(12)   NOT NULL,     -- (must allow duplicates)
   rid        integer unsigned NOT NULL,  -- recipient: maddr.id (dupl. allowed)
@@ -237,23 +260,24 @@ CREATE TABLE msgrcpt (
   wl         char(1)       DEFAULT ' ',  -- sender whitelisted by this recip
   bspam_level float,                     -- spam level + per-recip boost
   smtp_resp  varchar(255)  DEFAULT '',
-  KEY mail_id (mail_id),
-  KEY rid (rid)
+  FOREIGN KEY (rid)     REFERENCES maddr(id)     ON DELETE RESTRICT,
+  FOREIGN KEY (mail_id) REFERENCES msgs(mail_id) ON DELETE CASCADE
 ) ENGINE=InnoDB;
-CREATE INDEX msgrcpt_idx_mail_id ON msgrcpt (mail_id);
-CREATE INDEX msgrcpt_idx_rid     ON msgrcpt (rid);
+CREATE INDEX msgrcpt_idx_mail_id  ON msgrcpt (mail_id);
+CREATE INDEX msgrcpt_idx_rid      ON msgrcpt (rid);
 
 -- mail quarantine in SQL, enabled by $*_quarantine_method='sql:'
 -- NOTE: records in quarantine without corresponding msgs.mail_id record are
 --  orphaned and should be ignored and eventually deleted by external utilities
+-- NOTE: time_num field was introduced to this table with version 2.4.0
 CREATE TABLE quarantine (
   mail_id    varchar(12)   NOT NULL,    -- long-term unique mail id
-  chunk_ind  integer unsigned  NOT NULL,-- chunk number, starting with 1
+  chunk_ind  integer unsigned NOT NULL, -- chunk number, starting with 1
   mail_text  text          NOT NULL,    -- store mail as chunks up to 16 kB
-  PRIMARY KEY (mail_id,chunk_ind)
+  PRIMARY KEY (mail_id,chunk_ind),
+  FOREIGN KEY (mail_id) REFERENCES msgs(mail_id) ON DELETE CASCADE
 ) ENGINE=InnoDB;
 
-
 -- field msgrcpt.rs is primarily intended for use by quarantine management
 -- software; the value initially assigned by amavisd is a space;
 -- a short _preliminary_ list of possible values:
@@ -349,10 +373,11 @@ INSERT INTO wblist VALUES (17, 4, 'B');
 -- program are configurable through %sql_clause; see amavisd.conf-default
 
 
-An example of a query:
+
+Few examples of a query:
 
 SELECT
-  UNIX_TIMESTAMP()-time_num AS age, SUBSTRING(policy,1,2) as pb,
+  UNIX_TIMESTAMP()-msgs.time_num AS age, SUBSTRING(policy,1,2) as pb,
   content AS c, dsn_sent as dsn, ds, bspam_level AS level, size,
   SUBSTRING(sender.email,1,18) AS s,
   SUBSTRING(recip.email,1,18)  AS r,
@@ -360,24 +385,103 @@ SELECT
   FROM msgs LEFT JOIN msgrcpt         ON msgs.mail_id=msgrcpt.mail_id
             LEFT JOIN maddr AS sender ON msgs.sid=sender.id
             LEFT JOIN maddr AS recip  ON msgrcpt.rid=recip.id
-  WHERE content IS NOT NULL AND UNIX_TIMESTAMP()-time_num < 100
+  WHERE content IS NOT NULL AND UNIX_TIMESTAMP()-msgs.time_num < 100
   ORDER BY msgs.time_num DESC LIMIT 15;
 
+# clean messages ordered by count, grouped by domain:
+SELECT count(*) as cnt, avg(bspam_level), sender.domain
+  FROM msgs
+  LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
+  LEFT JOIN maddr AS sender ON msgs.sid=sender.id
+  LEFT JOIN maddr AS recip ON msgrcpt.rid=recip.id
+  WHERE content='C'
+  GROUP BY sender.domain ORDER BY cnt DESC LIMIT 50;
+
+# most spamy domains, sorted by spam average, grouped by domain:
+SELECT count(*) as cnt, avg(bspam_level) as spam_avg, sender.domain
+  FROM msgs
+  LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
+  LEFT JOIN maddr AS sender ON msgs.sid=sender.id
+  LEFT JOIN maddr AS recip ON msgrcpt.rid=recip.id
+  GROUP BY sender.domain ORDER BY spam_avg DESC LIMIT 50;
+
+# sender domains with >100 messages, sorted on sender.domain:
+SELECT count(*) as cnt, avg(bspam_level) as spam_avg, sender.domain
+  FROM msgs
+  LEFT JOIN msgrcpt ON msgs.mail_id=msgrcpt.mail_id
+  LEFT JOIN maddr AS sender ON msgs.sid=sender.id
+  LEFT JOIN maddr AS recip ON msgrcpt.rid=recip.id
+  GROUP BY sender.domain HAVING cnt > 100
+  ORDER BY sender.domain DESC LIMIT 100;
+
+
+
+
+Upgrading from pre 2.4.0 amavisd-new SQL schema to the 2.4.0 schema requires
+adding column 'quar_loc' to table msgs, and creating FOREIGN KEY constraint
+to facilitate deletion of expired records.
+
+The following clauses should be executed for upgrading pre-2.4.0 amavisd-new
+SQL schema to the 2.4.0 schema:
+
+  ALTER TABLE msgs ADD quar_loc varchar(255) DEFAULT '';
+
+  ALTER TABLE quarantine
+    ADD FOREIGN KEY (mail_id) REFERENCES msgs(mail_id) ON DELETE CASCADE;
+  ALTER TABLE msgrcpt
+    ADD FOREIGN KEY (mail_id) REFERENCES msgs(mail_id) ON DELETE CASCADE;
+
+--ALTER TABLE msgs
+--  ADD FOREIGN KEY (sid) REFERENCES maddr(id) ON DELETE RESTRICT;
+--ALTER TABLE msgrcpt
+--  ADD FOREIGN KEY (rid) REFERENCES maddr(id) ON DELETE RESTRICT;
+
 
 An example of a log/report/quarantine database housekeeping;
-delete old or orphaned records:
 
-DELETE FROM msgs WHERE UNIX_TIMESTAMP()-time_num > 7*24*60*60;
-DELETE FROM msgs WHERE UNIX_TIMESTAMP()-time_num > 60*60 AND content IS NULL;
+Instead of purging based on numerical Unix timestamp, one may select records
+based on ISO 8601 timestamps. This is particularly suitable for PostgreSQL:
+  DELETE FROM msgs WHERE time_iso < now() - interval '7 days';
+  DELETE FROM msgs WHERE time_iso < now() - interval '1 h' AND content IS NULL;
+and is also possible with MySQL, using slightly different format:
+  DELETE FROM msgs WHERE time_iso < now() - interval 7 day;
+  DELETE FROM msgs WHERE time_iso < now() -interval 1 hour AND content IS NULL;
 
-DELETE quarantine FROM quarantine LEFT JOIN msgs USING(mail_id)
-  WHERE msgs.mail_id IS NULL;
 
-DELETE msgrcpt    FROM msgrcpt    LEFT JOIN msgs USING(mail_id)
-  WHERE msgs.mail_id IS NULL;
 
+--  discarding indexes makes deletion faster; if we expect a larger proportion
+--  of records to be deleted it may be quicker to discard index, do deletions,
+--  and re-create index; for daily maintenance this does not pay off
+--DROP INDEX msgs_idx_sid         ON msgs;
+--DROP INDEX msgrcpt_idx_rid      ON msgrcpt;
+--DROP INDEX msgrcpt_idx_mail_id  ON msgrcpt;
+
+--  delete old msgs records based on timestamps only (for time_iso see above),
+--  and delete leftover msgs records from aborted mail checking operations
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-7*24*60*60;
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-60*60 AND content IS NULL;
+
+--  optionally certain content types may be given shorter lifetime
+DELETE FROM msgs WHERE time_num < UNIX_TIMESTAMP()-7*24*60*60
+  AND (content='V' OR (content='S' AND spam_level>20));
+
+--  (optional) just in case the ON DELETE CASCADE did not do its job, we may
+--  explicitly delete orphaned records (with no corresponding msgs entry);
+--  if ON DELETE CASCADE did work, there should be no deletions at this step
+DELETE FROM quarantine
+  WHERE NOT EXISTS (SELECT 1 FROM msgs WHERE mail_id=quarantine.mail_id);
+DELETE FROM msgrcpt
+  WHERE NOT EXISTS (SELECT 1 FROM msgs WHERE mail_id=msgrcpt.mail_id);
+
+--  re-create indexes (if they were removed in the first step):
+--CREATE INDEX msgs_idx_sid        ON msgs    (sid);
+--CREATE INDEX msgrcpt_idx_rid     ON msgrcpt (rid);
+--CREATE INDEX msgrcpt_idx_mail_id ON msgrcpt (mail_id);
+
+--  delete unreferenced e-mail addresses
 DELETE FROM maddr
-  WHERE NOT EXISTS (SELECT sid FROM msgs    WHERE sid=id)
-    AND NOT EXISTS (SELECT rid FROM msgrcpt WHERE rid=id);
+  WHERE NOT EXISTS (SELECT 1 FROM msgs    WHERE sid=id)
+    AND NOT EXISTS (SELECT 1 FROM msgrcpt WHERE rid=id);
 
-OPTIMIZE TABLE msgs, msgrcpt, maddr, quarantine;
+--  (optional) optimize tables from time to time
+--OPTIMIZE TABLE msgs, msgrcpt, quarantine, maddr;
